@@ -1,6 +1,6 @@
 ;;; context.el --- Support for ConTeXt documents.  -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2003-2021  Free Software Foundation, Inc.
+;; Copyright (C) 2003-2023  Free Software Foundation, Inc.
 
 ;; Maintainer: Berend de Boer <berend@pobox.com>
 ;; Keywords: tex
@@ -24,7 +24,7 @@
 
 ;;; Commentary:
 
-;; This is in progress ConTeXt support for AUCTeX. Please report
+;; This is in progress ConTeXt support for AUCTeX.  Please report
 ;; anomalies or things you believe should be added.
 
 ;; AUCTeX is closely interwoven with LaTeX.  We have to split up
@@ -37,10 +37,10 @@
 ;; 2. paragraph refilling doesn't work 100%, and is very slow.
 ;; 4. Remove dependency on LaTeX by moving LaTeX commands to TeX.
 ;; 5. Most ConTeXt macro's don't currently have lisp code to query for
-;;    arguments. As ConTeXt arguments are quite complex, the LaTeX way
+;;    arguments.  As ConTeXt arguments are quite complex, the LaTeX way
 ;;    of querying for arguments just doesn't cut it.
 ;; 6. Check auto-parsing: does it detect % interface=nl for example?
-;; 7. Complete adding ConTeXt macro's. Perhaps parse cont-en.xml and
+;; 7. Complete adding ConTeXt macro's.  Perhaps parse cont-en.xml and
 ;;    generate the interfaces?
 ;; 8. Add to menu: make TeX hash (mktexlsr), context format and metapost format.
 
@@ -49,7 +49,6 @@
 
 ;;; Code:
 
-(require 'tex-buf)
 (require 'tex)
 (require 'latex) ; for functions like `TeX-look-at' and `LaTeX-split-long-menu'
 (require 'plain-tex) ; for `plain-TeX-common-initialization'
@@ -161,7 +160,7 @@
   what)
 
 (defun ConTeXt-insert-other-macro (other-macro)
-  "Insert the ConTeXt other macro's macro SETUP."
+  "Insert the ConTeXt other macro's macro OTHER-MACRO."
   (insert TeX-esc (ConTeXt-other-macro-command other-macro))
   (newline)
   (indent-according-to-mode)
@@ -310,13 +309,20 @@ The following variables can be set to customize:
 
 ;; LaTeX has a max function here, which makes no sense.
 ;; I think you want to insert a section that is max ConTeXt-largest-level
+;; (May 3, 2023) The above comment is wrong.  Here "large" refers to
+;; coarseness of document structure grouping.  That is, "chapter" is
+;; larger than "section", "section" is larger than "subsection" etc.
+;; On the other hand, the corresponding levels are numbered in the
+;; reversed order.  That is, "chapter" is level 1, "section" is level
+;; 2 etc.  Hence the largest _section_ has the smallest _level_.
+;; That's the reason we use `max' rather than `min' here.
 (defun ConTeXt-current-section ()
   "Return the level of the section that contain point.
 See also `ConTeXt-section' for description of levels."
   (save-excursion
-    (min (ConTeXt-largest-level)
+    (max (ConTeXt-largest-level)
          (if (re-search-backward outline-regexp nil t)
-             (+ 1 (- (ConTeXt-outline-level) (ConTeXt-outline-offset)))
+             (- (ConTeXt-outline-level) (ConTeXt-outline-offset))
            (ConTeXt-largest-level)))))
 
 (defun ConTeXt-down-section ()
@@ -364,35 +370,19 @@ section."
 
 (defun ConTeXt-numbered-section-name (level)
   "Return the name of the section corresponding to LEVEL."
-  (let ((entry (TeX-member level ConTeXt-numbered-section-list
-                           (function (lambda (a b) (equal a (nth 1 b)))))))
-    (if entry
-        (nth 0 entry)
-      nil)))
+  (car (rassoc (list level) ConTeXt-numbered-section-list)))
 
 (defun ConTeXt-unnumbered-section-name (level)
   "Return the name of the section corresponding to LEVEL."
-  (let ((entry (TeX-member level ConTeXt-unnumbered-section-list
-                           (function (lambda (a b) (equal a (nth 1 b)))))))
-    (if entry
-        (nth 0 entry)
-      nil)))
+  (car (rassoc (list level) ConTeXt-unnumbered-section-list)))
 
 (defun ConTeXt-numbered-section-level (name)
   "Return the level of the section NAME."
-  (let ((entry (TeX-member name ConTeXt-numbered-section-list
-                           (function (lambda (a b) (equal a (nth 0 b)))))))
-    (if entry
-        (nth 1 entry)
-      nil)))
+  (cadr (assoc name ConTeXt-numbered-section-list)))
 
 (defun ConTeXt-unnumbered-section-level (name)
   "Return the level of the section NAME."
-  (let ((entry (TeX-member name ConTeXt-unnumbered-section-list
-                           (function (lambda (a b) (equal a (nth 0 b)))))))
-    (if entry
-        (nth 1 entry)
-      nil)))
+  (cadr (assoc name ConTeXt-unnumbered-section-list)))
 
 
 ;;; Section Hooks.
@@ -412,34 +402,36 @@ The following variables are set before the hooks are run
                       `ConTeXt-level'.
 `ConTeXt-title'     - The title of the section, default to an empty
                       string.
+`ConTeXt-reference' - Comma separated list of reference.
 `ConTeXt-done-mark' - Position of point afterwards, default nil
                       (meaning end).
 
-The following standard hook exist -
+The following standard hooks exist -
 
-ConTeXt-numbered-section-heading: Query the user about the name
+`ConTeXt-numbered-section-heading': Query the user about the name
 of the sectioning command.  Modifies `ConTeXt-level' and
 `ConTeXt-name'.
 
-ConTeXt-section-title: Query the user about the title of the
+`ConTeXt-section-title': Query the user about the title of the
 section.  Modifies `ConTeXt-title'.
 
-ConTeXt-section-section: Insert ConTeXt section command according
+`ConTeXt-section-section': Insert ConTeXt section command according
 to `ConTeXt-name', `ConTeXt-title', and `ConTeXt-reference'.  If
 `ConTeXt-title' is an empty string, `ConTeXt-done-mark' will be
 placed at the point they should be inserted.
 
-ConTeXt-section-ref: Insert a reference for this section command.
+`ConTeXt-section-ref': Query the user about a reference for this
+section command.  Modifies `ConTeXt-reference'.
 
 To get a full featured `ConTeXt-section' command, insert
 
  (setq ConTeXt-numbered-section-hook
-                         '(ConTeXt-numbered-section-heading
+                         \\='(ConTeXt-numbered-section-heading
                                  ConTeXt-section-title
-                                 ConTeXt-section-section
-                                 ConTeXt-section-ref))
+                                 ConTeXt-section-ref
+                                 ConTeXt-section-section))
 
-in your .emacs file."
+in your init file such as .emacs.d/init.el or .emacs."
   :group 'ConTeXt-macro
   :type 'hook
   :options
@@ -464,34 +456,36 @@ The following variables are set before the hooks are run
                       `ConTeXt-level'.
 `ConTeXt-title'     - The title of the section, default to an empty
                       string.
+`ConTeXt-reference' - Comma separated list of reference.
 `ConTeXt-done-mark' - Position of point afterwards, default nil
                       (meaning end).
 
-The following standard hook exist -
+The following standard hooks exist -
 
-ConTeXt-unnumbered-section-heading: Query the user about the name
+`ConTeXt-unnumbered-section-heading': Query the user about the name
 of the sectioning command.  Modifies `ConTeXt-level' and
 `ConTeXt-name'.
 
-ConTeXt-section-title: Query the user about the title of the
+`ConTeXt-section-title': Query the user about the title of the
 section.  Modifies `ConTeXt-title'.
 
-ConTeXt-section-section: Insert ConTeXt section command according
+`ConTeXt-section-section': Insert ConTeXt section command according
 to `ConTeXt-name', `ConTeXt-title', and `ConTeXt-reference'.  If
 `ConTeXt-title' is an empty string, `ConTeXt-done-mark' will be
 placed at the point they should be inserted.
 
-ConTeXt-section-ref: Insert a reference for this section command.
+`ConTeXt-section-ref': Query the user about a reference for this
+section command.  Modifies `ConTeXt-reference'.
 
 To get a full featured `ConTeXt-section' command, insert
 
  (setq ConTeXt-unnumbered-section-hook
-                         '(ConTeXt-unnumbered-section-heading
+                         \\='(ConTeXt-unnumbered-section-heading
                                  ConTeXt-section-title
-                                 ConTeXt-section-section
-                                 ConTeXt-section-ref))
+                                 ConTeXt-section-ref
+                                 ConTeXt-section-section))
 
-in your .emacs file."
+in your init file such as .emacs.d/init.el or .emacs."
   :group 'ConTeXt-macro
   :type 'hook
   :options
@@ -562,7 +556,7 @@ assumes the section already is inserted."
   (newline)
   ;; If RefTeX is available, tell it that we've just made a new section
   (and (fboundp 'reftex-notice-new-section)
-       (funcall (symbol-function 'reftex-notice-new-section))))
+       (reftex-notice-new-section)))
 
 (defun ConTeXt-section-ref ()
   "Hook to insert a reference after the sectioning command.
@@ -599,7 +593,7 @@ for a label to be inserted after the sectioning command."
     (cond ((TeX-TeX-sentinel-check process name))
           ((save-excursion
              ;; in a full ConTeXt run there will multiple texutil
-             ;; outputs. Just looking for "another run needed" would
+             ;; outputs.  Just looking for "another run needed" would
              ;; find the first occurence
              (goto-char (point-max))
              (re-search-backward "TeXUtil " nil t)
@@ -640,10 +634,7 @@ for a label to be inserted after the sectioning command."
 
 (TeX-auto-add-type "environment" "ConTeXt")
 
-(if (fboundp 'advice-add)               ;Emacs≥24.4 (or ELPA package nadvice)
-    (advice-add 'ConTeXt-add-environments :after #'ConTeXt--invalidate-menu)
-  (defadvice ConTeXt-add-environments (after ConTeXt-invalidate-menu (&rest environments) activate)
-    (ConTeXt--invalidate-menu)))
+(advice-add 'ConTeXt-add-environments :after #'ConTeXt--invalidate-menu)
 (defun ConTeXt--invalidate-menu (&rest _)
   "Mark the menu as being in need of a refresh."
   (setq ConTeXt-menu-changed t))
@@ -684,14 +675,14 @@ With optional ARG, modify current environment."
                    (t ConTeXt-default-environment)))
          (environment
           (completing-read (concat "Environment type (default " default "): ")
-                           ConTeXt-environment-list nil nil nil
+                           (ConTeXt-environment-list) nil nil nil
                            'ConTeXt-environment-history default)))
     ;; Use `environment' as default for the next time only if it is different
     ;; from the current default.
     (unless (equal environment default)
       (setq ConTeXt-default-environment environment))
 
-    (let ((entry (assoc environment ConTeXt-environment-list)))
+    (let ((entry (assoc environment (ConTeXt-environment-list))))
       (if (null entry)
           (ConTeXt-add-environments (list environment)))
 
@@ -700,7 +691,7 @@ With optional ARG, modify current environment."
         (ConTeXt-environment-menu environment)))))
 
 (defun ConTeXt-modify-environment (environment)
-  "Modify current environment."
+  "Modify current environment to new ENVIRONMENT."
   (save-excursion
     (ConTeXt-find-matching-stop)
     (re-search-backward (concat (regexp-quote TeX-esc)
@@ -721,7 +712,7 @@ With optional ARG, modify current environment."
 
 (defun ConTeXt-environment-menu (environment)
   "Insert ENVIRONMENT around point or region."
-  (let ((entry (assoc environment ConTeXt-environment-list)))
+  (let ((entry (assoc environment (ConTeXt-environment-list))))
     (cond ((not (and entry (nth 1 entry)))
            (ConTeXt-insert-environment environment))
           ((numberp (nth 1 entry))
@@ -731,16 +722,19 @@ With optional ARG, modify current environment."
                (setq args (concat args TeX-grop TeX-grcl))
                (setq count (- count 1)))
              (ConTeXt-insert-environment environment args)))
-          ((stringp (nth 1 entry))
+          ((or (stringp (nth 1 entry)) (vectorp (nth 1 entry)))
            (let ((prompts (cdr entry))
                  (args ""))
-             (while prompts
-               (setq args (concat args
-                                  TeX-grop
-                                  (read-from-minibuffer
-                                   (concat (car prompts) ": "))
-                                  TeX-grcl))
-               (setq prompts (cdr prompts)))
+             (dolist (elt prompts)
+               (let* ((optional (vectorp elt))
+                      (elt (if optional (elt elt 0) elt))
+                      (arg (TeX-read-string
+                            (TeX-argument-prompt optional elt nil))))
+                 (setq args (concat args
+                                    (cond ((and optional (> (length arg) 0))
+                                           (concat ConTeXt-optop arg ConTeXt-optcl))
+                                          ((not optional)
+                                           (concat TeX-grop arg TeX-grcl)))))))
              (ConTeXt-insert-environment environment args)))
           (t
            (apply (nth 1 entry) environment (nthcdr 2 entry))))))
@@ -760,45 +754,156 @@ With optional ARG, modify current environment."
   (end-of-line)
   (newline))
 
+(defvar ConTeXt-after-insert-env-hook nil
+  "List of functions to be run at the end of `ConTeXt-insert-environment'.
+Each function is called with three arguments: the name of the
+environment just inserted, the buffer position just before
+\\start... and the position just before \\stop....")
+
+;;; Copy and adaptation of `LaTeX-insert-environment'.  (2022-08-13)
 (defun ConTeXt-insert-environment (environment &optional extra)
-  "Insert ENVIRONMENT, with optional argument EXTRA."
-  (if (and (TeX-active-mark)
-           (not (eq (mark) (point))))
+  "Insert ConTeXt ENVIRONMENT with optional argument EXTRA."
+  (let ((active-mark (and (TeX-active-mark) (not (eq (mark) (point)))))
+        content-start env-start env-end additional-indent)
+    (when (and active-mark (< (mark) (point))) (exchange-point-and-mark))
+    ;; What to do with the line containing point.
+    ;; - Open a new empty line for later insertion of "\startfoo" and
+    ;;   put the point there.
+    ;; - If there were at first any non-whitespace texts between the
+    ;;   point and EOL, send them into their new own line.
+    (cond (;; When the entire line consists of whitespaces...
+           (save-excursion (beginning-of-line)
+                           (looking-at "[ \t]*$"))
+           ;; ...make the line empty and put the point there.
+           (delete-region (match-beginning 0) (match-end 0)))
+          (;; When there are only whitespaces between the point and
+           ;; BOL (including the case the point is at BOL)...
+           (TeX-looking-at-backward "^[ \t]*"
+                                    (line-beginning-position))
+           ;; ...in this case, we have non-whitespace texts between
+           ;; the point and EOL, so send the entire line into a new
+           ;; next line and put the point on the empty line just
+           ;; created.
+           (beginning-of-line)
+           (newline)
+           (beginning-of-line 0)
+           ;; Take note that there are texts to be indented later
+           ;; unless the region is activated.
+           (unless active-mark
+             (setq additional-indent t)))
+          (;; In all other cases...
+           t
+           ;; ...insert a new empty line after deleting all
+           ;; whitespaces around the point, put the point there...
+           (delete-horizontal-space)
+           (if (eolp)
+               (newline)
+             ;; ...and if there were at first any non-whitespace texts
+             ;; between (the original position of) the point and EOL,
+             ;; send them into a new next line.
+             (newline 2)
+             (beginning-of-line 0)
+             ;; Take note that there are texts to be indented later
+             ;; unless the region is activated.
+             (unless active-mark
+               (setq additional-indent t)))))
+    ;; What to do with the line containing mark.
+    ;; If there is active region...
+    (when active-mark
+      ;; - Open a new empty line for later insertion of "\stopfoo"
+      ;;   and put the mark there.
+      ;; - If there were at first any non-whitespace texts between the
+      ;;   mark and EOL, pass them over the empty line and put them on
+      ;;   their own line.
       (save-excursion
-        (if (< (mark) (point))
-            (exchange-point-and-mark))
-        (insert TeX-esc (ConTeXt-environment-start-name) environment)
-        (newline)
-        (forward-line -1)
-        (indent-according-to-mode)
-        (if extra (insert extra))
         (goto-char (mark))
-        (or (TeX-looking-at-backward "^[ \t]*")
-            (newline))
-        (insert TeX-esc (ConTeXt-environment-stop-name) environment)
-        (newline)
-        (forward-line -1)
-        (indent-according-to-mode)
-        ;;(goto-char (point))
-        )
-    (or (TeX-looking-at-backward "^[ \t]*")
-        (newline))
+        (cond (;; When the entire line consists of whitespaces...
+               (save-excursion (beginning-of-line)
+                               (looking-at "[ \t]*$"))
+               ;; ...make the line empty and put the mark there.
+               (delete-region (match-beginning 0) (match-end 0)))
+              (;; When there are only whitespaces between the mark and
+               ;; BOL (including the case the mark is at BOL)...
+               (TeX-looking-at-backward "^[ \t]*"
+                                        (line-beginning-position))
+               ;; ...in this case, we have non-whitespace texts
+               ;; between the mark and EOL, so send the entire line
+               ;; into a new next line and put the mark on the empty
+               ;; line just created.
+               (beginning-of-line)
+               (set-mark (point))
+               (newline)
+               ;; Take note that there are texts to be indented later.
+               (setq additional-indent t))
+              (;; In all other cases...
+               t
+               ;; ...make a new empty line after deleting all
+               ;; whitespaces around the mark, put the mark there...
+               (delete-horizontal-space)
+               (insert-before-markers "\n")
+               ;; ...and if there were at first any non-whitespace
+               ;; texts between (the original position of) the mark
+               ;; and EOL, send them into a new next line.
+               (unless (eolp)
+                 (newline)
+                 ;; Take note that there are texts to be indented
+                 ;; later.
+                 (setq additional-indent t))))))
+    ;; Now insert the environment.
+    (setq env-start (point))
     (insert TeX-esc (ConTeXt-environment-start-name) environment)
     (indent-according-to-mode)
-    (if extra (insert extra))
-    (end-of-line)
-    (newline-and-indent)
-    (newline)
+    (when extra (insert extra))
+    (setq content-start (line-beginning-position 2))
+    (unless active-mark
+      (newline)
+      (newline))
+    (when active-mark (goto-char (mark)))
     (insert TeX-esc (ConTeXt-environment-stop-name) environment)
-    (or (looking-at "[ \t]*$")
-        (save-excursion (newline-and-indent)))
-    (indent-according-to-mode)
-    (end-of-line 0)))
+    (end-of-line 0)
+    (if active-mark
+        (progn
+          ;; TODO: Do filling when context.el obtains
+          ;; `ConTeXt-fill-region' in future.
+          (indent-region content-start (line-beginning-position 2))
+          (set-mark content-start))
+      (indent-according-to-mode))
+    ;; Indent \stopfoo.
+    (save-excursion (beginning-of-line 2) (indent-according-to-mode)
+                    (when additional-indent
+                      ;; Indent texts sent after the inserted
+                      ;; environment.
+                      (forward-line 1) (indent-according-to-mode)))
+    (setq env-end (save-excursion
+                    (search-forward
+                     (concat TeX-esc (ConTeXt-environment-stop-name)
+                             environment))
+                    (match-beginning 0)))
+    (run-hook-with-args 'ConTeXt-after-insert-env-hook
+                        environment env-start env-end)))
+
+(defun ConTeXt--env-parse-args (args)
+  "Helper function to insert arguments defined by ARGS.
+This function checks if `TeX-exit-mark' is set, otherwise it's
+set to the point where this function starts.  Point will be at
+`TeX-exit-mark' when this function exits."
+  (let ((TeX-exit-mark (or TeX-exit-mark
+                           (point-marker))))
+    (ConTeXt-find-matching-start)
+    (end-of-line)
+    (TeX-parse-arguments args)
+    (goto-char TeX-exit-mark)
+    (set-marker TeX-exit-mark nil)))
+
+(defun ConTeXt-env-args (environment &rest args)
+  "Insert ENVIRONMENT and arguments defined by ARGS."
+  (ConTeXt-insert-environment environment)
+  (ConTeXt--env-parse-args args))
 
 
-;; with the following we can call a function on an environment. Say
+;; with the following we can call a function on an environment.  Say
 ;; you have metapost stuff within your TeX file, go to the environment
-;; and run ConTeXt-work-on-environment (suggested Key: C-c !). AUCTeX
+;; and run ConTeXt-work-on-environment (suggested Key: C-c !).  AUCTeX
 ;; sees that you are inside e.g. \startMPpage....\stopMPpage and
 ;; looks in ConTeXt-environment-helper for a function to be called.
 
@@ -843,7 +948,7 @@ An entry looks like: (\"environment\" . function)")
   (context-mode)
   (widen))
 
-;; find smarter name. Suggestions welcome
+;; find smarter name.  Suggestions welcome
 (defun ConTeXt-work-on-environment ()
   "Takes current environment and does something on it (todo: documentation)."
   (interactive)
@@ -939,10 +1044,10 @@ If INNER is non-nil, go to the point just past the \\start... macro."
                        (setq level (1- level))))))
     ;; now we have to look if we want to start behind the \start... macro
     (when inner
-      ;; \startfoo can have 0 or more {} and [] pairs. I assume that
-      ;; skipping all those parens will be smart enough. It fails when
+      ;; \startfoo can have 0 or more {} and [] pairs.  I assume that
+      ;; skipping all those parens will be smart enough.  It fails when
       ;; the first part in the \start-\stop-environment is { or [, like
-      ;; in \startquotation   {\em important} \stopquotation. There is
+      ;; in \startquotation   {\em important} \stopquotation.  There is
       ;; yet another pitfall: \startsetups SomeSetup foo bar
       ;; \stopsetups will use SomeSetup as the argument and the
       ;; environment
@@ -1061,7 +1166,8 @@ header is at the start of a line."
    "\\|" (ConTeXt-header-end) "\\b"
    "\\|" (ConTeXt-trailer-start) "\\b"))
 
-(defvar ConTeXt-text "Name of ConTeXt macro that begins the text body.")
+(defvar ConTeXt-text nil
+  "Name of ConTeXt macro that begins the text body.")
 
 (defun ConTeXt-header-end ()
   "Default end of header marker for ConTeXt documents."
@@ -1148,11 +1254,11 @@ An optional fourth (or sixth) element means always replace if t."
 (defun ConTeXt-outline-name ()
   "Guess a name for the current header line."
   (save-excursion
-    (if (re-search-forward "{\\([^\}]*\\)}" (point-at-eol) t)
+    (if (re-search-forward "{\\([^}]*\\)}" (line-end-position) t)
         (match-string 1)
-      (buffer-substring-no-properties (point) (point-at-eol)))))
+      (buffer-substring-no-properties (point) (line-end-position)))))
 
-;; This imenu also includes commented out chapters. Perhaps a feature
+;; This imenu also includes commented out chapters.  Perhaps a feature
 ;; for LaTeX, not sure we want or need that for ConTeXt.
 
 (defun ConTeXt-imenu-create-index-function ()
@@ -1318,8 +1424,10 @@ else.  There might be text before point."
   (let ((map (make-sparse-keymap)))
     (set-keymap-parent map TeX-mode-map)
 
-    (define-key map "\e\C-a"  #'ConTeXt-find-matching-start)
-    (define-key map "\e\C-e"  #'ConTeXt-find-matching-stop)
+    ;; We now set `beginning-of-defun-function' and
+    ;; `end-of-defun-function' instead.
+    ;; (define-key map "\e\C-a"  #'ConTeXt-find-matching-start)
+    ;; (define-key map "\e\C-e"  #'ConTeXt-find-matching-stop)
     ;; likely to change in the future
     (define-key map "\C-c!"    #'ConTeXt-work-on-environment)
     (define-key map "\C-c\C-e" #'ConTeXt-environment)
@@ -1351,20 +1459,20 @@ else.  There might be text before point."
 
 (defun ConTeXt-environment-menu-entry (entry)
   "Create an ENTRY for the environment menu."
-  (vector (car entry) (list 'ConTeXt-environment-menu (car entry)) t))
+  (vector (car entry) (list #'ConTeXt-environment-menu (car entry)) t))
 
 (defvar ConTeXt-environment-modify-menu-name "Change Environment   (C-u C-c C-e)")
 
 (defun ConTeXt-environment-modify-menu-entry (entry)
   "Create an ENTRY for the change environment menu."
-  (vector (car entry) (list 'ConTeXt-modify-environment (car entry)) t))
+  (vector (car entry) (list #'ConTeXt-modify-environment (car entry)) t))
 
 ;; ConTeXt define macros
 (defvar ConTeXt-define-menu-name "Define")
 
 (defun ConTeXt-define-menu-entry (entry)
   "Create an ENTRY for the define menu."
-  (vector entry (list 'ConTeXt-define-menu entry)))
+  (vector entry (list #'ConTeXt-define-menu entry)))
 
 (defun ConTeXt-define-menu (define)
   "Insert DEFINE from menu."
@@ -1375,7 +1483,7 @@ else.  There might be text before point."
 
 (defun ConTeXt-setup-menu-entry (entry)
   "Create an ENTRY for the setup menu."
-  (vector entry (list 'ConTeXt-setup-menu entry)))
+  (vector entry (list #'ConTeXt-setup-menu entry)))
 
 (defun ConTeXt-setup-menu (setup)
   "Insert SETUP from menu."
@@ -1386,7 +1494,7 @@ else.  There might be text before point."
 
 (defun ConTeXt-referencing-menu-entry (entry)
   "Create an ENTRY for the referencing menu."
-  (vector entry (list 'ConTeXt-referencing-menu entry)))
+  (vector entry (list #'ConTeXt-referencing-menu entry)))
 
 (defun ConTeXt-referencing-menu (referencing)
   "Insert REFERENCING from menu."
@@ -1397,10 +1505,10 @@ else.  There might be text before point."
 
 (defun ConTeXt-other-macro-menu-entry (entry)
   "Create an ENTRY for the other macro menu."
-  (vector entry (list 'ConTeXt-other-macro-menu entry)))
+  (vector entry (list #'ConTeXt-other-macro-menu entry)))
 
 (defun ConTeXt-other-macro-menu (other-macro)
-  "Insert OTHER MACRO from menu."
+  "Insert OTHER-MACRO from menu."
   (ConTeXt-insert-other-macro other-macro))
 
 
@@ -1416,7 +1524,7 @@ else.  There might be text before point."
 
 (defun ConTeXt-project-structure-menu-entry (entry)
   "Create an ENTRY for the project structure menu."
-  (vector entry (list 'ConTeXt-project-structure-menu entry)))
+  (vector entry (list #'ConTeXt-project-structure-menu entry)))
 
 
 ;; meta-structure section blocks menu entries
@@ -1429,7 +1537,7 @@ else.  There might be text before point."
 
 (defun ConTeXt-section-block-menu-entry (entry)
   "Create an ENTRY for the section block menu."
-  (vector entry (list 'ConTeXt-section-block-menu entry)))
+  (vector entry (list #'ConTeXt-section-block-menu entry)))
 
 
 ;; section menu entries
@@ -1464,25 +1572,23 @@ else.  There might be text before point."
 (defun ConTeXt-numbered-section-menu-entry (entry)
   "Create an ENTRY for the numbered section menu."
   (let ((enable (ConTeXt-section-enable-symbol (nth 1 entry))))
-    (set enable t)
-    (vector (car entry) (list 'ConTeXt-numbered-section-menu (nth 1 entry)) enable)))
+    (vector (car entry) (list #'ConTeXt-numbered-section-menu (nth 1 entry)) enable)))
 
 (defun ConTeXt-unnumbered-section-menu-entry (entry)
   "Create an ENTRY for the unnumbered section menu."
   (let ((enable (ConTeXt-section-enable-symbol (nth 1 entry))))
-    (set enable t)
-    (vector (car entry) (list 'ConTeXt-unnumbered-section-menu (nth 1 entry)) enable)))
+    (vector (car entry) (list #'ConTeXt-unnumbered-section-menu (nth 1 entry)) enable)))
 
 
 ;; etexshow support
 
 (defun ConTeXt-etexshow ()
-  "Call etexshow, if available, to show the definition of a ConText macro."
+  "Call etexshow, if available, to show the definition of a ConTeXt macro."
   (interactive)
-  (if (fboundp 'etexshow)
-      (let ()
+  (if (fboundp 'etexshow-cmd)
+      (progn
         (require 'etexshow)
-        (funcall (symbol-function 'etexshow-cmd)))
+        (etexshow-cmd))
     (error "etexshow is not installed.  Get it from http://levana.de/emacs/")))
 
 ;; menu itself
@@ -1653,7 +1759,7 @@ Use `ConTeXt-Mark-version' to choose the command."
      (let ((engine (eval (nth 4 (TeX-engine-in-engine-alist TeX-engine)) t)))
        (when engine
          (format "--engine=%s " engine)))
-     (unless (eq ConTeXt-current-interface "en")
+     (unless (string= ConTeXt-current-interface "en")
        (format "--interface=%s " ConTeXt-current-interface))
      (when TeX-source-correlate-mode
        (format "--passon=\"%s\" "
@@ -1683,22 +1789,25 @@ Use `ConTeXt-Mark-version' to choose the command."
     ConTeXt-extra-paragraph-commands
     ConTeXt-environment-list)
   "List of variables to be set from languages specific ones.")
+;; Make language specific variables buffer local
+(dolist (symbol ConTeXt-language-variable-list)
+  (make-variable-buffer-local symbol))
 
 (defconst ConTeXt-dialect :context
   "Default dialect for use with function `TeX-add-style-hook' for
 argument DIALECT-EXPR when the hook is to be run only on ConTeXt
-file, or any mode derived thereof. See variable
+file, or any mode derived thereof.  See variable
 `TeX-style-hook-dialect'." )
 
 (defcustom ConTeXt-clean-intermediate-suffixes
   ;; See *suffixes in texutil.pl.
   '("\\.tui" "\\.tup" "\\.ted" "\\.tes" "\\.top" "\\.log" "\\.tmp" "\\.run"
     "\\.bck" "\\.rlg" "\\.mpt" "\\.mpx" "\\.mpd" "\\.mpo" "\\.tuo" "\\.tub"
-    "\\.top" "-mpgraph\\.mp" "-mpgraph\\.mpd" "-mpgraph\\.mpo" "-mpgraph\\.mpy"
-    "-mprun\\.mp" "-mprun\\.mpd" "-mprun\\.mpo" "-mprun\\.mpy")
+    "\\.top" "\\.tuc" "-mpgraph\\.mp" "-mpgraph\\.mpd" "-mpgraph\\.mpo"
+    "-mpgraph\\.mpy" "-mprun\\.mp" "-mprun\\.mpd" "-mprun\\.mpo" "-mprun\\.mpy")
   "List of regexps matching suffixes of files to be deleted.
 The regexps will be anchored at the end of the file name to be matched,
-that is, you do _not_ have to cater for this yourself by adding \\\\' or $."
+that is, you do _not_ have to cater for this yourself by adding \\\\\\=' or $."
   :type '(repeat regexp)
   :group 'TeX-command)
 
@@ -1706,7 +1815,7 @@ that is, you do _not_ have to cater for this yourself by adding \\\\' or $."
   '("\\.dvi" "\\.pdf" "\\.ps")
   "List of regexps matching suffixes of files to be deleted.
 The regexps will be anchored at the end of the file name to be matched,
-that is, you do _not_ have to cater for this yourself by adding \\\\' or $."
+that is, you do _not_ have to cater for this yourself by adding \\\\\\=' or $."
   :type '(repeat regexp)
   :group 'TeX-command)
 
@@ -1726,20 +1835,16 @@ that is, you do _not_ have to cater for this yourself by adding \\\\' or $."
   (setq local-abbrev-table context-mode-abbrev-table)
   (set (make-local-variable 'TeX-style-hook-dialect) ConTeXt-dialect)
 
-  ;; Make language specific variables buffer local
-  (dolist (symbol ConTeXt-language-variable-list)
-    (make-variable-buffer-local symbol))
-
   (require (intern (concat "context-" ConTeXt-current-interface)))
   (dolist (symbol ConTeXt-language-variable-list)
     (set symbol (symbol-value (intern (concat (symbol-name symbol) "-"
                                               ConTeXt-current-interface)))))
 
-  ;; Create certain regular expressions based on language
-  (setq ConTeXt-indent-item-re (concat "\\\\\\(" (mapconcat #'identity ConTeXt-item-list "\\|") "\\)\\>"))
+  ;; Moved after `run-mode-hooks'. (bug#65750)
+  ;; (setq ConTeXt-indent-item-re (concat "\\\\\\(" (mapconcat #'identity ConTeXt-item-list "\\|") "\\)\\>"))
 
   ;; What's the deepest level at we can collapse a document?
-  ;; set only if user has not set it. Need to be set before menu is created.
+  ;; set only if user has not set it.  Need to be set before menu is created.
   ;; level 2 is "section"
   (or ConTeXt-largest-level
       (setq ConTeXt-largest-level 2))
@@ -1753,17 +1858,19 @@ that is, you do _not_ have to cater for this yourself by adding \\\\' or $."
 
   ;; Paragraph formatting
   (set (make-local-variable 'LaTeX-syntactic-comments) nil)
-  (set (make-local-variable 'LaTeX-paragraph-commands-regexp)
-       (ConTeXt-paragraph-commands-regexp))
+  ;; Moved after `run-mode-hooks'. (bug#65750)
+  ;; (set (make-local-variable 'LaTeX-paragraph-commands-regexp)
+  ;;      (ConTeXt-paragraph-commands-regexp))
   (set (make-local-variable 'paragraph-ignore-fill-prefix) t)
   (set (make-local-variable 'fill-paragraph-function) #'LaTeX-fill-paragraph)
   (set (make-local-variable 'adaptive-fill-mode) nil)
-  (setq paragraph-start
-        (concat
-         "[ \t]*\\("
-         (ConTeXt-paragraph-commands-regexp) "\\|"
-         "\\$\\$\\|" ; Plain TeX display math
-         "$\\)"))
+  ;; Moved after `run-mode-hooks'. (bug#65750)
+  ;; (setq paragraph-start
+  ;;       (concat
+  ;;        "[ \t]*\\("
+  ;;        (ConTeXt-paragraph-commands-regexp) "\\|"
+  ;;        "\\$\\$\\|" ; Plain TeX display math
+  ;;        "$\\)"))
   (setq paragraph-separate
         (concat
          "[ \t]*\\("
@@ -1774,12 +1881,17 @@ that is, you do _not_ have to cater for this yourself by adding \\\\' or $."
   (use-local-map ConTeXt-mode-map)
   (setq ConTeXt-menu-changed t)
 
+  ;; FIXME: Isn't `activate-menubar-hook' obsolete?
   (add-hook 'activate-menubar-hook #'ConTeXt-menu-update nil t)
 
+  (setq-local beginning-of-defun-function #'ConTeXt-find-matching-start)
+  (setq-local end-of-defun-function       #'ConTeXt-find-matching-stop)
+
   ;; Outline support
   (require 'outline)
-  (set (make-local-variable 'outline-level) 'ConTeXt-outline-level)
-  (set (make-local-variable 'outline-regexp) (ConTeXt-outline-regexp t))
+  (set (make-local-variable 'outline-level) #'ConTeXt-outline-level)
+  ;; Moved after `run-mode-hooks'. (bug#65750)
+  ;; (set (make-local-variable 'outline-regexp) (ConTeXt-outline-regexp t))
   ;;(make-local-variable 'outline-heading-end-regexp)
   (setq TeX-header-end (ConTeXt-header-end)
         TeX-trailer-start (ConTeXt-trailer-start))
@@ -1794,7 +1906,31 @@ that is, you do _not_ have to cater for this yourself by adding \\\\' or $."
   ;; run hooks
   (setq TeX-command-default "ConTeXt")
   (setq TeX-sentinel-default-function #'TeX-ConTeXt-sentinel)
-  (run-mode-hooks 'text-mode-hook 'TeX-mode-hook 'ConTeXt-mode-hook))
+  (run-mode-hooks 'text-mode-hook 'TeX-mode-hook 'ConTeXt-mode-hook)
+
+  ;; Create certain regular expressions based on language.
+  ;; Don't overwrite the value the user set by hooks or file
+  ;; (directory) variables.
+  (or (local-variable-p 'ConTeXt-indent-item-re)
+      (setq-local ConTeXt-indent-item-re
+                  (concat
+                   "\\\\\\("
+                   (mapconcat #'identity ConTeXt-item-list "\\|")
+                   "\\)\\>")))
+
+  ;; Don't do locally-bound test for `LaTeX-paragraph-commands-regexp'
+  ;; and `paragraph-start'.  See comments in similar part in latex.el.
+  (setq-local LaTeX-paragraph-commands-regexp
+              (ConTeXt-paragraph-commands-regexp))
+  (setq paragraph-start
+        (concat
+         "[ \t]*\\("
+         (ConTeXt-paragraph-commands-regexp) "\\|"
+         "\\$\\$\\|" ; Plain TeX display math
+         "$\\)"))
+
+  (or (local-variable-p 'outline-regexp)
+      (setq-local outline-regexp (ConTeXt-outline-regexp t))))
 
 (defun context-guess-current-interface ()
   "Guess what ConTeXt interface the current buffer is using."
