@@ -1,6 +1,6 @@
 ;;; subcaption.el --- AUCTeX style for `subcaption.sty' (v1.3)  -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2015--2021 Free Software Foundation, Inc.
+;; Copyright (C) 2015--2023 Free Software Foundation, Inc.
 
 ;; Author: Arash Esbati <arash@gnu.org>
 ;; Maintainer: auctex-devel@gnu.org
@@ -136,21 +136,31 @@ caption, insert only a caption."
    ;; The next 2 macros are part of the kernel of caption.sty, but we
    ;; load them within subcaption.el.
    (TeX-add-symbols
-    '("DeclareCaptionSubType"
-      [TeX-arg-eval
-       completing-read (TeX-argument-prompt t nil "Numbering scheme")
-       '("arabic" "roman" "Roman" "alph" "Alph" "fnsymbol")]
-      (TeX-arg-eval
-       completing-read (TeX-argument-prompt nil nil "Type")
-       '("figure" "table")))
+    `("DeclareCaptionSubType"
+      [TeX-arg-completing-read ("arabic" "roman" "Roman"
+                                "alph" "Alph" "fnsymbol")
+                               "Numbering scheme"]
+      (TeX-arg-completing-read
+       ,(lambda ()
+          (append
+           (when (and (fboundp 'LaTeX-newfloat-DeclareFloatingEnvironment-list)
+                      (LaTeX-newfloat-DeclareFloatingEnvironment-list))
+             (mapcar #'car (LaTeX-newfloat-DeclareFloatingEnvironment-list)))
+           '("figure" "table")))
+       "Type"))
 
-    '("DeclareCaptionSubType*"
-      [TeX-arg-eval completing-read
-                    (TeX-argument-prompt t nil "Numbering scheme")
-                    '("arabic" "roman" "Roman" "alph" "Alph" "fnsymbol")]
-      (TeX-arg-eval completing-read
-                    (TeX-argument-prompt nil nil "Type")
-                    '("figure" "table"))))
+    `("DeclareCaptionSubType*"
+      [TeX-arg-completing-read ("arabic" "roman" "Roman"
+                                "alph" "Alph" "fnsymbol")
+                               "Numbering scheme"]
+      (TeX-arg-completing-read
+       ,(lambda ()
+          (append
+           (when (and (fboundp 'LaTeX-newfloat-DeclareFloatingEnvironment-list)
+                      (LaTeX-newfloat-DeclareFloatingEnvironment-list))
+             (mapcar #'car (LaTeX-newfloat-DeclareFloatingEnvironment-list)))
+           '("figure" "table")))
+       "Type")))
 
    ;; \subcaption(box)? and \subfloat macros should get their own lines
    (LaTeX-paragraph-commands-add-locally
@@ -186,10 +196,15 @@ caption, insert only a caption."
                               'function)) )
  TeX-dialect)
 
+(defvar LaTeX-subcaption-package-options-list
+  (progn
+    (TeX-load-style "caption")
+    (append LaTeX-subcaption-key-val-options
+            LaTeX-caption-key-val-options))
+  "Package options for the subcaption package.")
+
 (defun LaTeX-subcaption-package-options ()
   "Prompt for package options for the subcaption package."
-  (TeX-load-style "caption")
-  (TeX-read-key-val t (append LaTeX-subcaption-key-val-options
-                              LaTeX-caption-key-val-options)))
+  (TeX-read-key-val t LaTeX-subcaption-package-options-list))
 
 ;;; subcaption.el ends here
