@@ -1,6 +1,6 @@
 ;;; tex-fold.el --- Fold TeX macros.  -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2004-2021  Free Software Foundation, Inc.
+;; Copyright (C) 2004-2023  Free Software Foundation, Inc.
 
 ;; Author: Ralf Angeli <angeli@caeruleus.net>
 ;; Maintainer: auctex-devel@gnu.org
@@ -62,8 +62,8 @@
 
 (defcustom TeX-fold-type-list '(env macro math)
   "List of item types to consider when folding.
-Valid items are the symbols 'env for environments, 'macro for
-macros, 'math for math macros and 'comment for comments."
+Valid items are the symbols `env' for environments, `macro' for
+macros, `math' for math macros and `comment' for comments."
   :type '(set (const :tag "Environments" env)
               (const :tag "Macros" macro)
               (const :tag "Math Macros" math)
@@ -73,7 +73,7 @@ macros, 'math for math macros and 'comment for comments."
   '(("[f]" ("footnote" "marginpar"))
     ("[c]" ("cite"))
     ("[l]" ("label"))
-    ("[r]" ("ref" "pageref" "eqref"))
+    ("[r]" ("ref" "pageref" "eqref" "footref"))
     ("[i]" ("index" "glossary"))
     ("[1]:||*" ("item"))
     ("..." ("dots"))
@@ -111,6 +111,8 @@ the respective macro argument.
 If the first element is a function symbol, the function will be
 called with all mandatory arguments of the macro and the result
 of the function call will be used as a replacement for the macro.
+Such functions typically return a string, but may also return the
+symbol `abort' to indicate that the macro should not be folded.
 
 Setting this variable does not take effect immediately.  Use
 Customize or reset the mode."
@@ -123,7 +125,7 @@ Customize or reset the mode."
   "Internal list of display strings and macros to fold.
 Is updated when the TeX Fold mode is being activated and then
 contains all constructs to fold for the given buffer or mode
-respectively, i.e. contents of both `TeX-fold-macro-spec-list'
+respectively, that is, contents of both `TeX-fold-macro-spec-list'
 and <mode-prefix>-fold-macro-spec-list.")
 (make-variable-buffer-local 'TeX-fold-macro-spec-list-internal)
 
@@ -131,28 +133,30 @@ and <mode-prefix>-fold-macro-spec-list.")
   '(("[comment]" ("comment")))
   "List of display strings and environments to fold."
   :type '(repeat (group (choice (string :tag "Display String")
-                                (integer :tag "Number of argument" :value 1))
+                                (integer :tag "Number of argument" :value 1)
+                                (function :tag "Function to execute"))
                         (repeat :tag "Environments" (string)))))
 
 (defvar TeX-fold-env-spec-list-internal nil
   "Internal list of display strings and environments to fold.
 Is updated when the TeX Fold mode is being activated and then
 contains all constructs to fold for the given buffer or mode
-respectively, i.e. contents of both `TeX-fold-env-spec-list'
+respectively, that is, contents of both `TeX-fold-env-spec-list'
 and <mode-prefix>-fold-env-spec-list.")
 (make-variable-buffer-local 'TeX-fold-env-spec-list-internal)
 
 (defcustom TeX-fold-math-spec-list nil
   "List of display strings and math macros to fold."
   :type '(repeat (group (choice (string :tag "Display String")
-                                (integer :tag "Number of argument" :value 1))
+                                (integer :tag "Number of argument" :value 1)
+                                (function :tag "Function to execute"))
                         (repeat :tag "Math Macros" (string)))))
 
 (defvar TeX-fold-math-spec-list-internal nil
   "Internal list of display strings and math macros to fold.
 Is updated when the TeX Fold mode is being activated and then
 contains all constructs to fold for the given buffer or mode
-respectively, i.e. contents of both `TeX-fold-math-spec-list'
+respectively, that is, contents of both `TeX-fold-math-spec-list'
 and <mode-prefix>-fold-math-spec-list.")
 (make-variable-buffer-local 'TeX-fold-math-spec-list-internal)
 
@@ -311,8 +315,8 @@ and `TeX-fold-math-spec-list', and environments in `TeX-fold-env-spec-list'."
 
 (defun TeX-fold-region-macro-or-env (start end type)
   "Fold all items of type TYPE in region from START to END.
-TYPE can be one of the symbols 'env for environments, 'macro
-for macros and 'math for math macros."
+TYPE can be one of the symbols `env' for environments, `macro'
+for macros and `math' for math macros."
   (save-excursion
     (let (fold-list item-list regexp)
       (dolist (item (cond ((eq type 'env) TeX-fold-env-spec-list-internal)
@@ -414,7 +418,7 @@ for macros and 'math for math macros."
 (defun TeX-fold-item (type)
   "Hide the item on which point currently is located.
 TYPE specifies the type of item and can be one of the symbols
-'env for environments, 'macro for macros or 'math for math
+`env' for environments, `macro' for macros or `math' for math
 macros.
 Return non-nil if an item was found and folded, nil otherwise."
   (if (and (eq type 'env)
@@ -508,11 +512,11 @@ Return non-nil if a comment was found and folded, nil otherwise."
 (defun TeX-fold-make-overlay (ov-start ov-end type display-string-spec)
   "Make a TeX-fold overlay extending from OV-START to OV-END.
 TYPE is a symbol which is used to describe the content to hide
-and may be 'macro for macros, 'math for math macro and 'env for
+and may be `macro' for macros, `math' for math macro and `env' for
 environments.
 DISPLAY-STRING-SPEC is the original specification of the display
-string in the variables `TeX-fold-macro-spec-list' or
-`TeX-fold-env-spec-list' and may be a string or an integer."
+string in the variables `TeX-fold-macro-spec-list' and alikes.
+See its doc string for detail."
   ;; Calculate priority before the overlay is instantiated.  We don't
   ;; want `TeX-overlay-prioritize' to pick up a non-prioritized one.
   (let ((priority (TeX-overlay-prioritize ov-start ov-end))
@@ -526,8 +530,8 @@ string in the variables `TeX-fold-macro-spec-list' or
 
 (defun TeX-fold-item-end (start type)
   "Return the end of an item of type TYPE starting at START.
-TYPE can be either 'env for environments, 'macro for macros or
-'math for math macros."
+TYPE can be either `env' for environments, `macro' for macros or
+`math' for math macros."
   (save-excursion
     (cond ((and (eq type 'env)
                 (eq major-mode 'context-mode))
@@ -745,7 +749,8 @@ Replace them with the respective macro argument."
         (setq spec elt
               index nil)
         ;; Find and expand every placeholder.
-        (while (and (string-match "\\([[{<]\\)\\([1-9]\\)\\([]}>]\\)" elt index)
+        (while (and (string-match "\\([[{<(]\\)\\([1-9]\\)\\([]}>)]\\)"
+                                  elt index)
                     ;; Does the closing delim match the opening one?
                     (string-equal
                      (match-string 3 elt)
@@ -794,25 +799,40 @@ That means, put respective properties onto overlay OV."
          (display-string (if (listp computed) (car computed) computed))
          ;; (face (when (listp computed) (cadr computed)))
          )
-    ;; Do nothing if the overlay is empty.
-    (when (and ov-start ov-end)
-      ;; Cater for zero-length display strings.
-      (when (string= display-string "") (setq display-string TeX-fold-ellipsis))
-      ;; Add a linebreak to the display string and adjust the overlay end
-      ;; in case of an overfull line.
-      (when (TeX-fold-overfull-p ov-start ov-end display-string)
-        (setq display-string (concat display-string "\n"))
-        (move-overlay ov ov-start (save-excursion
-                                    (goto-char ov-end)
-                                    (skip-chars-forward " \t")
-                                    (point))))
-      (overlay-put ov 'mouse-face 'highlight)
-      (overlay-put ov 'display display-string)
-      (when font-lock-mode
-        (overlay-put ov 'face TeX-fold-folded-face))
-      (unless (zerop TeX-fold-help-echo-max-length)
-        (overlay-put ov 'help-echo (TeX-fold-make-help-echo
-                                    (overlay-start ov) (overlay-end ov)))))))
+
+    (if (eq computed 'abort)
+        ;; Abort folding if computed result is the symbol `abort'.
+        ;; This allows programmatic customization.
+        ;; Suggested by Paul Nelson <ultrono@gmail.com>.
+        ;; <URL:https://lists.gnu.org/r/auctex/2023-08/msg00026.html>
+        (progn (delete-overlay ov)
+               t ; so that `TeX-fold-dwim' "gives up"
+               )
+      ;; Do nothing if the overlay is empty.
+      (when (and ov-start ov-end)
+        ;; Cater for zero-length display strings.
+        (when (string= display-string "") (setq display-string TeX-fold-ellipsis))
+        ;; Add a linebreak to the display string and adjust the overlay end
+        ;; in case of an overfull line.
+        (when (TeX-fold-overfull-p ov-start ov-end display-string)
+          (setq display-string (concat display-string "\n"))
+          (move-overlay ov ov-start (save-excursion
+                                      (goto-char ov-end)
+                                      (skip-chars-forward " \t")
+                                      (point))))
+        (overlay-put ov 'mouse-face 'highlight)
+        (when font-lock-mode
+          ;; Add raise adjustment for superscript and subscript.
+          ;; (bug#42209)
+          (setq display-string
+                (propertize display-string
+                            'display (get-text-property ov-start 'display))))
+        (overlay-put ov 'display display-string)
+        (when font-lock-mode
+          (overlay-put ov 'face TeX-fold-folded-face))
+        (unless (zerop TeX-fold-help-echo-max-length)
+          (overlay-put ov 'help-echo (TeX-fold-make-help-echo
+                                      (overlay-start ov) (overlay-end ov))))))))
 
 (defun TeX-fold-show-item (ov)
   "Show a single LaTeX macro or environment.
@@ -897,10 +917,16 @@ the other elements.  The ordering among elements is maintained."
 Called interactively, with no prefix argument, toggle the mode.
 With universal prefix ARG (or if ARG is nil) turn mode on.
 With zero or negative ARG turn mode off."
-  nil nil (list (cons TeX-fold-command-prefix TeX-fold-keymap))
+  :init-value nil
+  :lighter nil
+  :keymap (list (cons TeX-fold-command-prefix TeX-fold-keymap))
   (if TeX-fold-mode
       (progn
-        (set (make-local-variable 'search-invisible) t)
+        ;; The value t causes problem when body text is hidden in
+        ;; outline-minor-mode. (bug#36651)
+        ;; In addition, it's better not to override user preference
+        ;; without good reason.
+        ;; (set (make-local-variable 'search-invisible) t)
         (add-hook 'post-command-hook #'TeX-fold-post-command nil t)
         (add-hook 'LaTeX-fill-newline-hook #'TeX-fold-update-at-point nil t)
         (add-hook 'TeX-after-insert-macro-hook
@@ -922,7 +948,7 @@ With zero or negative ARG turn mode off."
                                                      (TeX-mode-prefix) elt))))
                          (when (boundp symbol)
                            (symbol-value symbol)))))))
-    (kill-local-variable 'search-invisible)
+    ;; (kill-local-variable 'search-invisible)
     (remove-hook 'post-command-hook #'TeX-fold-post-command t)
     (remove-hook 'LaTeX-fill-newline-hook #'TeX-fold-update-at-point t)
     (TeX-fold-clearout-buffer))
